@@ -1,3 +1,5 @@
+# $Id: PicStart.pm,v 1.6 2005/08/27 18:22:10 kissg Exp $
+
 ###############################################################################
 # Frontend
 #
@@ -19,7 +21,7 @@ sub new {
 
 	my %args = @_;
 	my $self = Pista::Programmer::PicStart::BackEnd->
-				new($line, $args{debug}) or return;
+								new($line, $args{debug}) or return;
 	$self->{progress} = $args{progress}; # Flag: Progress report required.
 	undef %args;
 
@@ -92,8 +94,8 @@ sub write_prog {
 		$chunksize = $mtu if $chunksize > $mtu;
 		my @sublist = splice(@values, 0, $chunksize);
 		$status = $self->SUPER::write_prog($device,
-						$start, $start+$chunksize-1,
-						\@sublist);
+										$start, $start+$chunksize-1,
+										\@sublist);
 		$status or return;
 		$start += $chunksize;
 		$len -= $chunksize;
@@ -112,7 +114,7 @@ sub read_prog {
 		my $chunksize = $len;
 		$chunksize = $mtu if $chunksize > $mtu;
 		my $content = $self->SUPER::read_prog($device,
-						$start, $start+$chunksize-1);
+										$start, $start+$chunksize-1);
 		$content or return;
 		push(@retval, @$content);
 		$start += $chunksize;
@@ -149,7 +151,8 @@ sub read_devid {
 		# 0x01 0x02 0x03 0x04 0x08 0x09 0x0c 0x0d 0x0f 0x10 0x11
 		$device = dup($device);
 		$device->{userid}->{len} = $end - $device->{userid}->{addr} + 1;
-		my $devidref = $self->read_userid($device, $device->{userid}->{addr}, $end);
+		my $devidref = $self->read_userid($device,
+									$device->{userid}->{addr}, $end);
 		splice(@$devidref, 0, $start - $device->{userid}->{addr});
 		return $devidref;
 	}
@@ -171,7 +174,7 @@ use warnings;
 use Pista::Util qw(dup);
 our @ISA = ();		# initial value
 
-# 18Fxxx devices are written in 8 byte blocks however original PS+ protocol
+# Newer devices are written in multiword blocks however original PS+ protocol
 # expect single words to write then returns verified word.
 # Therefore verifying is just simulated during write operation.
 # The actual verification must be done separately.
@@ -180,8 +183,8 @@ sub write_prog {
 	my ($self, $device, $start, $end, $values) = @_;
 	$self->check_firmware($device) or return;
 	my $status = $self->SUPER::write_prog($device, $start, $end, $values);
-	return $status unless $status;				# write failed
-	return $status if $device->{picstart}->{arch} != 0x0e;	# verified
+	return $status unless $status;							# write failed
+	return $status if $device->{prog}->{blocksize} == 1;	# verified
 	my $content = $self->read_prog($device, $start, $end);
 	return $self->_verify($device, 'prog', $values, $content);
 }
@@ -190,8 +193,8 @@ sub write_userid {
 	my ($self, $device, $start, $end, $values) = @_;
 	$self->check_firmware($device) or return;
 	my $status = $self->SUPER::write_userid($device, $start, $end, $values);
-	return $status unless $status;				# write failed
-	return $status if $device->{picstart}->{arch} != 0x0e;	# verified
+	return $status unless $status;							# write failed
+	return $status if $device->{userid}->{blocksize} == 1;	# verified
 	my $content = $self->read_userid($device, $start, $end);
 	return $self->_verify($device, 'userid', $values, $content);
 }
@@ -241,8 +244,8 @@ sub _protect {
 	my @zeroconf;
 	if (exists $device->{conf}->{preserve}) {
 		@zeroconf = @{ $self->read_conf($device,
-						$device->{conf}->{addr}, 
-						$device->{conf}->{end}) };
+									$device->{conf}->{addr}, 
+									$device->{conf}->{end}) };
 		for my $i (0..$#{$device->{conf}->{blank}}) {
 			$zeroconf[$i] &= $device->{conf}->{preserve}->[$i];
 		}
@@ -284,11 +287,11 @@ sub blankcheck {
 		unless $EEblankcheck[$device->{picstart}->{arch}] eq 'Buggy';
 	$retval->{eeprom} = 1;
 	my $content = $self->SUPER::read_eeprom($device,
-			$device->{eeprom}->{addr},
-			$device->{eeprom}->{addr} + $device->{eeprom}->{len}-1);
+					$device->{eeprom}->{addr},
+					$device->{eeprom}->{addr} + $device->{eeprom}->{len}-1);
 	for (my $i=0; $i<=$#{$content}; $i++) {
 		next if $device->{eeprom}->{blank}->[0] ==
-			$content->[$i] & $device->{eeprom}->{mask}->[0];
+			($content->[$i] & $device->{eeprom}->{mask}->[0]);
 		return $retval;
 	}
 	$retval->{eeprom} = 0;
@@ -310,42 +313,42 @@ use Pista::Util qw(dup CCtov vtoCC);
 
 use constant SPEED			=> 19200;
 
-use constant BLANK_CHECK_FULL		=> 0x42;
-use constant DIAGNOSTICS		=> 0x43;
-use constant BLANK_CHECK_RANGE		=> 0x44;
-use constant WRITE_PROGRAM		=> 0x51;
-use constant READ_PROGRAM		=> 0x54;
-use constant READ_CAL			=> 0x63;
-use constant READ_EEPROM		=> 0x64;
-use constant READ_USERID		=> 0x65;
-use constant READ_CONFIG		=> 0x66;
-use constant WRITE_CONFIG		=> 0x67;
-use constant WRITE_USERID		=> 0x68;
-use constant WRITE_EEPROM		=> 0x69;
-use constant WRITE_CAL			=> 0x71;
-use constant GET_INFO_LEN		=> 0x80;
-use constant LOAD_PROC_INFO		=> 0x81;
-use constant LOAD_EXTENDED_CONFIG	=> 0x82;
-use constant DOWNLOAD_FIRMWARE		=> 0x87;	# 4.0.0+
-use constant FW_MAGIC			=> 0x55aa;	# 4.0.0+
-use constant ASK_HARDWARE		=> 0x88;
-use constant  PICSTART_PLUS		=> 0xab;
-use constant GET_VERSION		=> 0x8d;
-use constant SET_RANGE			=> 0x8e;
-use constant ERASE			=> 0x8f;
-use constant COMPUTE_PROGRAM_CHKSUM	=> 0x90;
-use constant COMPUTE_CONF_CHKSUM	=> 0x91;
-use constant COMPUTE_USERID_CHKSUM	=> 0x92;
-use constant COMPUTE_EEPROM_CHKSUM	=> 0x93;
-use constant COMPUTE_CAL_CHKSUM		=> 0x94;
-use constant READ_DEVID			=> undef;	# unsupported by 4.10.6
+use constant BLANK_CHECK_FULL			=> 0x42;
+use constant DIAGNOSTICS				=> 0x43;
+use constant BLANK_CHECK_RANGE			=> 0x44;
+use constant WRITE_PROGRAM				=> 0x51;
+use constant READ_PROGRAM				=> 0x54;
+use constant READ_CAL					=> 0x63;
+use constant READ_EEPROM				=> 0x64;
+use constant READ_USERID				=> 0x65;
+use constant READ_CONFIG				=> 0x66;
+use constant WRITE_CONFIG				=> 0x67;
+use constant WRITE_USERID				=> 0x68;
+use constant WRITE_EEPROM				=> 0x69;
+use constant WRITE_CAL					=> 0x71;
+use constant GET_INFO_LEN				=> 0x80;
+use constant LOAD_PROC_INFO				=> 0x81;
+use constant LOAD_EXTENDED_CONFIG		=> 0x82;
+use constant DOWNLOAD_FIRMWARE			=> 0x87;	# 4.0.0+
+use constant FW_MAGIC					=> 0x55aa;	# 4.0.0+
+use constant ASK_HARDWARE				=> 0x88;
+use constant  PICSTART_PLUS				=> 0xab;
+use constant GET_VERSION				=> 0x8d;
+use constant SET_RANGE					=> 0x8e;
+use constant ERASE						=> 0x8f;
+use constant COMPUTE_PROGRAM_CHKSUM		=> 0x90;
+use constant COMPUTE_CONF_CHKSUM		=> 0x91;
+use constant COMPUTE_USERID_CHKSUM		=> 0x92;
+use constant COMPUTE_EEPROM_CHKSUM		=> 0x93;
+use constant COMPUTE_CAL_CHKSUM			=> 0x94;
+use constant READ_DEVID					=> undef;	# unsupported by 4.10.6
 
-use constant PROG_NOTBLANK		=> 0x01;
-use constant CONF_NOTBLANK		=> 0x02;
-use constant USERID_NOTBLANK		=> 0x04;
-use constant EEPROM_NOTBLANK		=> 0x08;
-use constant CAL_NOTBLANK		=> 0x10;
-use constant PARITY_ERROR		=> 0x20;
+use constant PROG_NOTBLANK				=> 0x01;
+use constant CONF_NOTBLANK				=> 0x02;
+use constant USERID_NOTBLANK			=> 0x04;
+use constant EEPROM_NOTBLANK			=> 0x08;
+use constant CAL_NOTBLANK				=> 0x10;
+use constant PARITY_ERROR				=> 0x20;
 
 BEGIN {
 	eval "sub CRTSCTS { 020000000000; }" # POSIX.pm does not define this
@@ -382,16 +385,16 @@ sub new {
 	$termios->setcflag(CS8 | HUPCL | CREAD | CLOCAL | CRTSCTS);
 	$termios->setispeed(B0);
 	$termios->setospeed(B0);
-	$termios->setattr($fd, TCSANOW);	# DTR down: reset PICSTART Plus
+	$termios->setattr($fd, TCSANOW);		# DTR down: reset PICSTART Plus
 	print "120 Resetting programmer\n";
 	sleep 1;
 	my $posixspeed = eval "B".SPEED;
 	$termios->setispeed($posixspeed);
 	$termios->setospeed($posixspeed);
-	$termios->setattr($fd, TCSANOW);	# DTR up
+	$termios->setattr($fd, TCSANOW);		# DTR up
 
 	select((select($self->{fh}),$|=1)[0]);	# make handle unbuffered
-	tcflush($fd, TCIFLUSH);			# discard any pending input
+	tcflush($fd, TCIFLUSH);					# discard any pending input
 
 	# Identify programmer
 	$self->_send_serial('C', ASK_HARDWARE);
@@ -470,14 +473,18 @@ sub read_conf {
 	$self->_download_processor_info($device) or return;
 	my $content = $self->_read(READ_CONFIG, $len, 2);
 	$content = vtoCC($content) if ($device->{conf}->{width} == 1);
+	pop(@$content) if $#$content > $end-$start;
 	return $content;
 }
 
 sub write_conf {
 	my ($self, $device, $start, $end, $values) = @_;
+	my $masks = [];
 	$self->check_firmware($device) or return;
+#$dumper->dumpValue($device);
 	my $relstart = $start - $device->{conf}->{addr};
 	for my $i (0..$#{$values}) {
+		$masks->[$i] = $device->{conf}->{mask}->[$relstart+$i];
 		next if defined $values->[$i];
 		$values->[$i] = $device->{conf}->{blank}->[$relstart+$i];
 	}
@@ -491,7 +498,7 @@ sub write_conf {
 	$sect->{end} = $end;
 	$sect->{len} = $end - $start + 1;
 	$self->_download_processor_info($device) or return;
-	return $self->_write(WRITE_CONFIG, $len, 2, $values);
+	return $self->_write(WRITE_CONFIG, $len, 2, $values, $masks);
 }
 
 sub blankcheck {
@@ -506,7 +513,7 @@ sub blankcheck {
 		conf	=> $status & CONF_NOTBLANK,
 		userid	=> $status & USERID_NOTBLANK,
 		eeprom	=> $status & EEPROM_NOTBLANK,
-		cal	=> $status & CAL_NOTBLANK,
+		cal		=> $status & CAL_NOTBLANK,
 	};
 }
 
@@ -519,6 +526,7 @@ sub read_prog {
 	$self->_set_range($start, $len) or return;
 	my $content = $self->_read(READ_PROGRAM, $len, 2);
 	$content = vtoCC($content) if ($device->{prog}->{width} == 1);
+	pop(@$content) if $#$content > $end-$start;
 	return $content;
 }
 
@@ -683,6 +691,7 @@ sub read_cal {
 	$self->_download_processor_info($device) or return;
 	my $content = $self->_read(READ_CAL, $len, 2);
 	$content = vtoCC($content) if ($device->{cal}->{width} == 1);
+	pop(@$content) if $#$content > $end-$start;
 	return $content;
 }
 
@@ -878,7 +887,7 @@ sub write_firmware {
 }
 
 sub _write {
-	my ($self, $wrcmd, $len, $width, $values) = @_;
+	my ($self, $wrcmd, $len, $width, $values, $masks) = @_;
 	if ($len < 0) {
 		die "Internal program error: len == $len";
 	}
@@ -906,7 +915,12 @@ sub _write {
 			print "524 Timeout\n";
 			return;
 		}
-		$errors++ if ($echo != $values->[$i]);
+		if (defined $masks) {
+			$errors++ if ($echo&$masks->[$i]) != ($values->[$i]&$masks->[$i]);
+		}
+		else {
+			$errors++ if $echo != $values->[$i];
+		}
 		next unless $self->{progress};
 		next unless $len > 255;
 		next unless time - $lastprogress >= 2;
@@ -1154,7 +1168,7 @@ sub _words {
 	my ($start, $end, $width) = @_;
 	my $len = $end - $start + 1;
 	return $len if ($width == 2);
-	die "Internal program error: odd length\n" if ($len % 2);
+	$len += $len%2;
 	return $len/2;
 }
 
@@ -1188,3 +1202,5 @@ sub check_firmware {
 }
 
 1;
+
+# vi: ts=4 sw=4
